@@ -1,32 +1,25 @@
-# coding=utf-8
-import sys
 import re
+from collections.abc import Mapping
 
 import pytest
 
 from tinydb import TinyDB, where, Query
 from tinydb.middlewares import Middleware, CachingMiddleware
-from tinydb.storages import MemoryStorage
-
-try:
-    import ujson as json
-except ImportError:
-    HAS_UJSON = False
-else:
-    HAS_UJSON = True
+from tinydb.storages import MemoryStorage, JSONStorage
+from tinydb.table import Document
 
 
-def test_purge(db):
-    db.purge()
+def test_drop_tables(db: TinyDB):
+    db.drop_tables()
 
     db.insert({})
-    db.purge()
+    db.drop_tables()
 
     assert len(db) == 0
 
 
-def test_all(db):
-    db.purge()
+def test_all(db: TinyDB):
+    db.drop_tables()
 
     for i in range(10):
         db.insert({})
@@ -34,13 +27,13 @@ def test_all(db):
     assert len(db.all()) == 10
 
 
-def test_insert(db):
-    db.purge()
+def test_insert(db: TinyDB):
+    db.drop_tables()
     db.insert({'int': 1, 'char': 'a'})
 
     assert db.count(where('int') == 1) == 1
 
-    db.purge()
+    db.drop_tables()
 
     db.insert({'int': 1, 'char': 'a'})
     db.insert({'int': 1, 'char': 'b'})
@@ -50,14 +43,30 @@ def test_insert(db):
     assert db.count(where('char') == 'a') == 1
 
 
-def test_insert_ids(db):
-    db.purge()
+def test_insert_ids(db: TinyDB):
+    db.drop_tables()
     assert db.insert({'int': 1, 'char': 'a'}) == 1
     assert db.insert({'int': 1, 'char': 'a'}) == 2
 
 
-def test_insert_multiple(db):
-    db.purge()
+def test_insert_with_doc_id(db: TinyDB):
+    db.drop_tables()
+    assert db.insert({'int': 1, 'char': 'a'}) == 1
+    assert db.insert(Document({'int': 1, 'char': 'a'}, 12)) == 12
+    assert db.insert(Document({'int': 1, 'char': 'a'}, 77)) == 77
+    assert db.insert({'int': 1, 'char': 'a'}) == 78
+
+
+def test_insert_with_duplicate_doc_id(db: TinyDB):
+    db.drop_tables()
+    assert db.insert({'int': 1, 'char': 'a'}) == 1
+
+    with pytest.raises(ValueError):
+        db.insert(Document({'int': 1, 'char': 'a'}, 1))
+
+
+def test_insert_multiple(db: TinyDB):
+    db.drop_tables()
     assert not db.contains(where('int') == 1)
 
     # Insert multiple from list
@@ -73,7 +82,7 @@ def test_insert_multiple(db):
         for j in range(10):
             yield {'int': j}
 
-    db.purge()
+    db.drop_tables()
 
     db.insert_multiple(generator())
 
@@ -82,7 +91,7 @@ def test_insert_multiple(db):
     assert db.count(where('int').exists()) == 10
 
     # Insert multiple from inline generator
-    db.purge()
+    db.drop_tables()
 
     db.insert_multiple({'int': i} for i in range(10))
 
@@ -90,8 +99,8 @@ def test_insert_multiple(db):
         assert db.count(where('int') == i) == 1
 
 
-def test_insert_multiple_with_ids(db):
-    db.purge()
+def test_insert_multiple_with_ids(db: TinyDB):
+    db.drop_tables()
 
     # Insert multiple from list
     assert db.insert_multiple([{'int': 1, 'char': 'a'},
@@ -99,14 +108,27 @@ def test_insert_multiple_with_ids(db):
                                {'int': 1, 'char': 'c'}]) == [1, 2, 3]
 
 
-def test_insert_invalid_type_raises_error(db):
+def test_insert_multiple_with_doc_ids(db: TinyDB):
+    db.drop_tables()
+
+    assert db.insert_multiple([
+        Document({'int': 1, 'char': 'a'}, 12),
+        Document({'int': 1, 'char': 'b'}, 77)
+    ]) == [12, 77]
+    assert db.get(doc_id=12) == {'int': 1, 'char': 'a'}
+    assert db.get(doc_id=77) == {'int': 1, 'char': 'b'}
+
+    with pytest.raises(ValueError):
+        db.insert_multiple([Document({'int': 1, 'char': 'a'}, 12)])
+
+
+def test_insert_invalid_type_raises_error(db: TinyDB):
     with pytest.raises(ValueError, match='Document is not a Mapping'):
-        db.insert(object())  # object() as an example of a non-mapping-type
+        # object() as an example of a non-mapping-type
+        db.insert(object())  # type: ignore
 
 
-def test_insert_valid_mapping_type(db):
-    from tinydb.database import Mapping
-
+def test_insert_valid_mapping_type(db: TinyDB):
     class CustomDocument(Mapping):
         def __init__(self, data):
             self.data = data
@@ -120,14 +142,12 @@ def test_insert_valid_mapping_type(db):
         def __len__(self):
             return len(self.data)
 
-    db.purge()
+    db.drop_tables()
     db.insert(CustomDocument({'int': 1, 'char': 'a'}))
     assert db.count(where('int') == 1) == 1
 
 
-def test_cutom_mapping_type_with_json(tmpdir):
-    from tinydb.database import Mapping
-
+def test_custom_mapping_type_with_json(tmpdir):
     class CustomDocument(Mapping):
         def __init__(self, data):
             self.data = data
@@ -143,7 +163,7 @@ def test_cutom_mapping_type_with_json(tmpdir):
 
     # Insert
     db = TinyDB(str(tmpdir.join('test.db')))
-    db.purge()
+    db.drop_tables()
     db.insert(CustomDocument({'int': 1, 'char': 'a'}))
     assert db.count(where('int') == 1) == 1
 
@@ -158,40 +178,40 @@ def test_cutom_mapping_type_with_json(tmpdir):
 
     # Write back
     doc_id = db.get(where('int') == 3).doc_id
-    db.write_back([CustomDocument({'int': 4, 'char': 'a'})], [doc_id])
+    db.update(CustomDocument({'int': 4, 'char': 'a'}), doc_ids=[doc_id])
     assert db.count(where('int') == 3) == 0
     assert db.count(where('int') == 4) == 1
 
 
-def test_remove(db):
+def test_remove(db: TinyDB):
     db.remove(where('char') == 'b')
 
     assert len(db) == 2
     assert db.count(where('int') == 1) == 2
 
 
-def test_remove_all_fails(db):
+def test_remove_all_fails(db: TinyDB):
     with pytest.raises(RuntimeError):
         db.remove()
 
 
-def test_remove_multiple(db):
+def test_remove_multiple(db: TinyDB):
     db.remove(where('int') == 1)
 
     assert len(db) == 0
 
 
-def test_remove_ids(db):
+def test_remove_ids(db: TinyDB):
     db.remove(doc_ids=[1, 2])
 
     assert len(db) == 1
 
 
-def test_remove_returns_ids(db):
+def test_remove_returns_ids(db: TinyDB):
     assert db.remove(where('char') == 'b') == [2]
 
 
-def test_update(db):
+def test_update(db: TinyDB):
     assert len(db) == 3
 
     db.update({'int': 2}, where('char') == 'a')
@@ -200,23 +220,23 @@ def test_update(db):
     assert db.count(where('int') == 1) == 2
 
 
-def test_update_all(db):
+def test_update_all(db: TinyDB):
     assert db.count(where('int') == 1) == 3
 
     db.update({'newField': True})
 
-    assert db.count(where('newField') == True) == 3
+    assert db.count(where('newField') == True) == 3  # noqa
 
 
-def test_update_returns_ids(db):
-    db.purge()
+def test_update_returns_ids(db: TinyDB):
+    db.drop_tables()
     assert db.insert({'int': 1, 'char': 'a'}) == 1
     assert db.insert({'int': 1, 'char': 'a'}) == 2
 
     assert db.update({'char': 'b'}, where('int') == 1) == [1, 2]
 
 
-def test_update_transform(db):
+def test_update_transform(db: TinyDB):
     def increment(field):
         def transform(el):
             el[field] += 1
@@ -239,60 +259,43 @@ def test_update_transform(db):
     assert db.count(where('int') == 1) == 2
 
 
-def test_update_ids(db):
+def test_update_ids(db: TinyDB):
     db.update({'int': 2}, doc_ids=[1, 2])
 
     assert db.count(where('int') == 2) == 2
 
 
-def test_write_back(db):
-    docs = db.search(where('int') == 1)
-    for doc in docs:
-        doc['int'] = [1, 2, 3]
+def test_update_multiple(db: TinyDB):
+    assert len(db) == 3
 
-    db.write_back(docs)
-    assert db.count(where('int') == [1, 2, 3]) == 3
+    db.update_multiple([
+        ({'int': 2}, where('char') == 'a'),
+        ({'int': 4}, where('char') == 'b'),
+    ])
+
+    assert db.count(where('int') == 1) == 1
+    assert db.count(where('int') == 2) == 1
+    assert db.count(where('int') == 4) == 1
 
 
-def test_write_back_whole_doc(db):
-    docs = db.search(where('int') == 1)
-    doc_ids = [doc.doc_id for doc in docs]
-    for i, doc in enumerate(docs):
-        docs[i] = {'newField': i}
+def test_update_multiple_operation(db: TinyDB):
+    def increment(field):
+        def transform(el):
+            el[field] += 1
 
-    db.write_back(docs, doc_ids)
-    assert db.count(where('newField') == 0) == 1
-    assert db.count(where('newField') == 1) == 1
-    assert db.count(where('newField') == 2) == 1
+        return transform
+
+    assert db.count(where('int') == 1) == 3
+
+    db.update_multiple([
+        (increment('int'), where('char') == 'a'),
+        (increment('int'), where('char') == 'b')
+    ])
+
+    assert db.count(where('int') == 2) == 2
 
 
-def test_write_back_returns_ids(db):
-    db.purge()
-    assert db.insert({'int': 1, 'char': 'a'}) == 1
-    assert db.insert({'int': 1, 'char': 'a'}) == 2
-
-    docs = [{'word': 'hello'}, {'word': 'world'}]
-
-    assert db.write_back(docs, [1, 2]) == [1, 2]
-
-
-def test_write_back_fails(db):
-    with pytest.raises(ValueError):
-        db.write_back([{'get': 'error'}], [1, 2])
-
-
-def test_write_back_id_exceed(db):
-    db.purge()
-    db.insert({'int': 1})
-    with pytest.raises(IndexError):
-        db.write_back([{'get': 'error'}], [2])
-
-
-def test_write_back_empty_ok(db):
-    db.write_back([])
-
-
-def test_upsert(db):
+def test_upsert(db: TinyDB):
     assert len(db) == 3
 
     # Document existing
@@ -304,7 +307,36 @@ def test_upsert(db):
     assert db.count(where('int') == 9) == 1
 
 
-def test_search(db):
+def test_upsert_by_id(db: TinyDB):
+    assert len(db) == 3
+
+    # Single document existing
+    extant_doc = Document({'char': 'v'}, doc_id=1)
+    assert db.upsert(extant_doc) == [1]
+    doc = db.get(where('char') == 'v')
+    assert isinstance(doc, Document)
+    assert doc is not None
+    assert doc.doc_id == 1
+    assert len(db) == 3
+
+    # Single document missing
+    missing_doc = Document({'int': 5, 'char': 'w'}, doc_id=5)
+    assert db.upsert(missing_doc) == [5]
+    doc = db.get(where('char') == 'w')
+    assert isinstance(doc, Document)
+    assert doc is not None
+    assert doc.doc_id == 5
+    assert len(db) == 4
+
+    # Missing doc_id and condition
+    with pytest.raises(ValueError, match=r"(?=.*\bdoc_id\b)(?=.*\bquery\b)"):
+        db.upsert({'no_Document': 'no_query'})
+
+    # Make sure we didn't break anything
+    assert db.insert({'check': '_next_id'}) == 6
+
+
+def test_search(db: TinyDB):
     assert not db._query_cache
     assert len(db.search(where('int') == 1)) == 3
 
@@ -312,47 +344,65 @@ def test_search(db):
     assert len(db.search(where('int') == 1)) == 3  # Query result from cache
 
 
-def test_search_path(db):
+def test_search_path(db: TinyDB):
     assert not db._query_cache
-    assert len(db.search(where('int'))) == 3
+    assert len(db.search(where('int').exists())) == 3
     assert len(db._query_cache) == 1
 
-    assert len(db.search(where('asd'))) == 0
-    assert len(db.search(where('int'))) == 3  # Query result from cache
+    assert len(db.search(where('asd').exists())) == 0
+    assert len(db.search(where('int').exists())) == 3  # Query result from cache
 
 
-def test_search_no_results_cache(db):
-    assert len(db.search(where('missing'))) == 0
-    assert len(db.search(where('missing'))) == 0
+def test_search_no_results_cache(db: TinyDB):
+    assert len(db.search(where('missing').exists())) == 0
+    assert len(db.search(where('missing').exists())) == 0
 
 
-def test_get(db):
+def test_get(db: TinyDB):
     item = db.get(where('char') == 'b')
+    assert isinstance(item, Document)
+    assert item is not None
     assert item['char'] == 'b'
 
 
-def test_get_ids(db):
+def test_get_ids(db: TinyDB):
     el = db.all()[0]
     assert db.get(doc_id=el.doc_id) == el
-    assert db.get(doc_id=float('NaN')) is None
+    assert db.get(doc_id=float('NaN')) is None  # type: ignore
 
 
-def test_count(db):
+def test_get_multiple_ids(db: TinyDB):
+    el = db.all()
+    assert db.get(doc_ids=[x.doc_id for x in el]) == el
+
+
+def test_get_invalid(db: TinyDB):
+    with pytest.raises(RuntimeError):
+        db.get()
+
+
+def test_count(db: TinyDB):
     assert db.count(where('int') == 1) == 3
     assert db.count(where('char') == 'd') == 0
 
 
-def test_contains(db):
+def test_contains(db: TinyDB):
     assert db.contains(where('int') == 1)
     assert not db.contains(where('int') == 0)
 
 
-def test_contains_ids(db):
-    assert db.contains(doc_ids=[1, 2])
-    assert not db.contains(doc_ids=[88])
+def test_contains_ids(db: TinyDB):
+    assert db.contains(doc_id=1)
+    assert db.contains(doc_id=2)
+    assert not db.contains(doc_id=88)
 
 
-def test_get_idempotent(db):
+def test_contains_invalid(db: TinyDB):
+    with pytest.raises(RuntimeError):
+        db.contains()
+
+
+def test_get_idempotent(db: TinyDB):
     u = db.get(where('int') == 1)
     z = db.get(where('int') == 1)
     assert u == z
@@ -376,7 +426,7 @@ def test_multiple_dbs():
 
 
 def test_storage_closed_once():
-    class Storage(object):
+    class Storage:
         def __init__(self):
             self.closed = False
 
@@ -418,7 +468,7 @@ def test_unique_ids(tmpdir):
 
     # Verify ids stay unique when inserting/removing
     with TinyDB(path) as _db:
-        _db.purge()
+        _db.drop_tables()
         _db.insert_multiple({'x': i} for i in range(5))
         _db.remove(where('x') == 2)
 
@@ -442,59 +492,7 @@ def test_lastid_after_open(tmpdir):
         _db.insert_multiple({'i': i} for i in range(NUM))
 
     with TinyDB(path) as _db:
-        assert _db._last_id == NUM
-
-
-@pytest.mark.skipif(sys.version_info >= (3, 0),
-                    reason="requires python2")
-def test_unicode_memory(db):
-    """
-    Regression test for issue #28
-    """
-    unic_str = 'ß'.decode('utf-8')
-    byte_str = 'ß'
-
-    db.insert({'value': unic_str})
-    assert db.contains(where('value') == byte_str)
-    assert db.contains(where('value') == unic_str)
-
-    db.purge()
-    db.insert({'value': byte_str})
-    assert db.contains(where('value') == byte_str)
-    assert db.contains(where('value') == unic_str)
-
-
-@pytest.mark.skipif(sys.version_info >= (3, 0),
-                    reason="requires python2")
-def test_unicode_json(tmpdir):
-    """
-    Regression test for issue #28
-    """
-    unic_str1 = 'a'.decode('utf-8')
-    byte_str1 = 'a'
-
-    unic_str2 = 'ß'.decode('utf-8')
-    byte_str2 = 'ß'
-
-    path = str(tmpdir.join('db.json'))
-
-    with TinyDB(path) as _db:
-        _db.purge()
-        _db.insert({'value': byte_str1})
-        _db.insert({'value': byte_str2})
-        assert _db.contains(where('value') == byte_str1)
-        assert _db.contains(where('value') == unic_str1)
-        assert _db.contains(where('value') == byte_str2)
-        assert _db.contains(where('value') == unic_str2)
-
-    with TinyDB(path) as _db:
-        _db.purge()
-        _db.insert({'value': unic_str1})
-        _db.insert({'value': unic_str2})
-        assert _db.contains(where('value') == byte_str1)
-        assert _db.contains(where('value') == unic_str1)
-        assert _db.contains(where('value') == byte_str2)
-        assert _db.contains(where('value') == unic_str2)
+        assert _db._get_next_id() - 1 == NUM
 
 
 def test_doc_ids_json(tmpdir):
@@ -505,17 +503,18 @@ def test_doc_ids_json(tmpdir):
     path = str(tmpdir.join('db.json'))
 
     with TinyDB(path) as _db:
-        _db.purge()
+        _db.drop_tables()
         assert _db.insert({'int': 1, 'char': 'a'}) == 1
         assert _db.insert({'int': 1, 'char': 'a'}) == 2
 
-        _db.purge()
+        _db.drop_tables()
         assert _db.insert_multiple([{'int': 1, 'char': 'a'},
                                     {'int': 1, 'char': 'b'},
                                     {'int': 1, 'char': 'c'}]) == [1, 2, 3]
 
-        assert _db.contains(doc_ids=[1, 2])
-        assert not _db.contains(doc_ids=[88])
+        assert _db.contains(doc_id=1)
+        assert _db.contains(doc_id=2)
+        assert not _db.contains(doc_id=88)
 
         _db.update({'int': 2}, doc_ids=[1, 2])
         assert _db.count(where('int') == 2) == 2
@@ -546,7 +545,6 @@ def test_insert_string(tmpdir):
         _db.insert({'int': 3})  # Does not fail
 
 
-@pytest.mark.skipif(HAS_UJSON, reason="not compatible with ujson")
 def test_insert_invalid_dict(tmpdir):
     path = str(tmpdir.join('db.json'))
 
@@ -574,51 +572,30 @@ def test_gc(tmpdir):
     db.close()
 
 
-def test_non_default_table():
+def test_drop_table():
     db = TinyDB(storage=MemoryStorage)
-    assert [TinyDB.DEFAULT_TABLE] == list(db.tables())
+    default_table_name = db.table(db.default_table_name).name
 
-    db = TinyDB(storage=MemoryStorage, default_table='non-default')
-    assert {'non-default'} == db.tables()
+    assert [] == list(db.tables())
+    db.drop_table(default_table_name)
 
-    db.purge_tables()
-    default_table = TinyDB.DEFAULT_TABLE
+    db.insert({'a': 1})
+    assert [default_table_name] == list(db.tables())
 
-    TinyDB.DEFAULT_TABLE = 'non-default'
-    db = TinyDB(storage=MemoryStorage)
-    assert {'non-default'} == db.tables()
-
-    TinyDB.DEFAULT_TABLE = default_table
-
-
-def test_non_default_table_args():
-    TinyDB.DEFAULT_TABLE_KWARGS = {'cache_size': 0}
-
-    db = TinyDB(storage=MemoryStorage)
-    default_table = db.table()
-    assert default_table._query_cache.capacity == 0
-
-    TinyDB.DEFAULT_TABLE_KWARGS = {}
-
-
-def test_purge_table():
-    db = TinyDB(storage=MemoryStorage)
-    assert [TinyDB.DEFAULT_TABLE] == list(db.tables())
-
-    db.purge_table(TinyDB.DEFAULT_TABLE)
+    db.drop_table(default_table_name)
     assert [] == list(db.tables())
 
     table_name = 'some-other-table'
     db = TinyDB(storage=MemoryStorage)
-    db.table(table_name)
-    assert {TinyDB.DEFAULT_TABLE, table_name} == db.tables()
+    db.table(table_name).insert({'a': 1})
+    assert {table_name} == db.tables()
 
-    db.purge_table(table_name)
-    assert {TinyDB.DEFAULT_TABLE} == db.tables()
-    assert table_name not in db._table_cache
+    db.drop_table(table_name)
+    assert set() == db.tables()
+    assert table_name not in db._tables
 
-    db.purge_table('non-existent-table-name')
-    assert {TinyDB.DEFAULT_TABLE} == db.tables()
+    db.drop_table('non-existent-table-name')
+    assert set() == db.tables()
 
 
 def test_empty_write(tmpdir):
@@ -629,7 +606,7 @@ def test_empty_write(tmpdir):
             raise AssertionError('No write for unchanged db')
 
     TinyDB(path).close()
-    TinyDB(path, storage=ReadOnlyMiddleware()).close()
+    TinyDB(path, storage=ReadOnlyMiddleware(JSONStorage)).close()
 
 
 def test_query_cache():
@@ -646,147 +623,30 @@ def test_query_cache():
 
     # Modify the db instance to not return any results when
     # bypassing the query cache
-    db._table_cache[TinyDB.DEFAULT_TABLE]._read = lambda: {}
+    db._tables[db.table(db.default_table_name).name]._read_table = lambda: {}
 
     # Make sure we got an independent copy of the result list
     results.extend([1])
     assert db.search(query) == [{'name': 'foo', 'value': 42}]
 
 
-def test_tinydb_is_iterable(db):
+def test_tinydb_is_iterable(db: TinyDB):
     assert [r for r in db] == db.all()
 
 
-def test_eids(db):
-    with pytest.warns(DeprecationWarning):
-        assert db.contains(eids=[1]) is True
-
-    with pytest.warns(DeprecationWarning):
-        db.update({'field': 'value'}, eids=[1])
-        assert db.contains(where('field') == 'value')
-
-    with pytest.warns(DeprecationWarning):
-        doc = db.get(eid=1)
-
-    with pytest.warns(DeprecationWarning):
-        assert doc.eid == 1
-
-    with pytest.warns(DeprecationWarning):
-        db.remove(eids=[1])
-        assert not db.contains(where('field') == 'value')
-
-    with pytest.raises(TypeError):
-        db.remove(eids=[1], doc_ids=[1])
-
-    with pytest.raises(TypeError):
-        db.get(eid=[1], doc_id=[1])
-
-
-def test_custom_table_class():
-    from tinydb.database import Table
-
-    class MyTableClass(Table):
-        pass
-
-    # Table class for single table
-    db = TinyDB(storage=MemoryStorage)
-    assert isinstance(TinyDB(storage=MemoryStorage).table(),
-                      Table)
-    assert isinstance(db.table('my_table', table_class=MyTableClass),
-                      MyTableClass)
-
-    # Table class for all tables
-    TinyDB.table_class = MyTableClass
-    assert isinstance(TinyDB(storage=MemoryStorage).table(),
-                      MyTableClass)
-    assert isinstance(TinyDB(storage=MemoryStorage).table('my_table'),
-                      MyTableClass)
-
-    # Reset default table class
-    TinyDB.table_class = Table
-
-
-def test_string_key():
-    from collections import Mapping
-
-    from tinydb.database import Table, StorageProxy, Document
-    from tinydb.storages import MemoryStorage
-
-    class StorageProxy2(StorageProxy):
-        def _new_document(self, key, val):
-            # Don't convert the key to a number here!
-            return Document(val, key)
-
-    class Table2(Table):
-        def _init_last_id(self, data):
-            if data:
-                self._last_id = len(data)
-            else:
-                self._last_id = 0
-
-        def _get_next_id(self):
-            next_id = self._last_id + 1
-            data = self._read()
-            while str(next_id) in data:
-                next_id += 1
-            self._last_id = next_id
-            return str(next_id)
-
-        def insert(self, document):
-            if not isinstance(document, Mapping):
-                raise ValueError('Document is not a Mapping')
-
-            doc_id = document.get('doc_id') or self._get_next_id()
-
-            data = self._read()
-            data[doc_id] = dict(document)
-            self._write(data)
-
-        def insert_multiple(self, documents):
-            """
-            Insert multiple documents into the table.
-
-            :param documents: a list of documents to insert
-            :returns: a list containing the inserted documents' IDs
-            """
-
-            doc_ids = []
-            data = self._read()
-
-            for document in documents:
-                if not isinstance(document, Mapping):
-                    raise ValueError('Document is not a Mapping')
-
-                doc_id = document.get('doc_id') or self._get_next_id()
-                doc_ids.append(doc_id)
-
-                data[doc_id] = dict(document)
-
-            self._write(data)
-
-            return doc_ids
-
-    db = TinyDB(storage=MemoryStorage, table_class=Table2,
-                storage_proxy_class=StorageProxy2)
-    table = db.table()
-    table.insert({'doc_id': 'abc'})
-    assert table.get(doc_id='abc')['doc_id'] == 'abc'
-    assert table._last_id == 0
-    table.insert({'abc': 10})
-    assert table.get(doc_id='1')['abc'] == 10
-    assert table._last_id == 1
-
-
 def test_repr(tmpdir):
     path = str(tmpdir.join('db.json'))
 
+    db = TinyDB(path)
+    db.insert({'a': 1})
+
     assert re.match(
         r"<TinyDB "
         r"tables=\[u?\'_default\'\], "
         r"tables_count=1, "
-        r"default_table_documents_count=0, "
-        r"all_tables_documents_count=\[\'_default=0\'\]>",
-        repr(TinyDB(path)))
+        r"default_table_documents_count=1, "
+        r"all_tables_documents_count=\[\'_default=1\'\]>",
+        repr(db))
 
 
 def test_delete(tmpdir):
@@ -804,10 +664,10 @@ def test_delete(tmpdir):
     assert db.search(q.network.id == '114') == []
 
 
-def test_insert_multiple_with_single_dict(db):
+def test_insert_multiple_with_single_dict(db: TinyDB):
     with pytest.raises(ValueError):
         d = {'first': 'John', 'last': 'smith'}
-        db.insert_multiple(d)
+        db.insert_multiple(d)  # type: ignore
         db.close()
 
 
@@ -816,3 +676,41 @@ def test_access_storage():
                       MemoryStorage)
     assert isinstance(TinyDB(storage=CachingMiddleware(MemoryStorage)).storage,
                       CachingMiddleware)
+
+
+def test_empty_db_len():
+    db = TinyDB(storage=MemoryStorage)
+    assert len(db) == 0
+
+
+def test_insert_on_existing_db(tmpdir):
+    path = str(tmpdir.join('db.json'))
+
+    db = TinyDB(path, ensure_ascii=False)
+    db.insert({'foo': 'bar'})
+
+    assert len(db) == 1
+
+    db.close()
+
+    db = TinyDB(path, ensure_ascii=False)
+    db.insert({'foo': 'bar'})
+    db.insert({'foo': 'bar'})
+
+    assert len(db) == 3
+
+
+def test_storage_access():
+    db = TinyDB(storage=MemoryStorage)
+
+    assert isinstance(db.storage, MemoryStorage)
+
+
+def test_lambda_query():
+    db = TinyDB(storage=MemoryStorage)
+    db.insert({'foo': 'bar'})
+
+    query = lambda doc: doc.get('foo') == 'bar'
+    query.is_cacheable = lambda: False
+    assert db.search(query) == [{'foo': 'bar'}]
+    assert not db._query_cache
