@@ -17,17 +17,41 @@ False
 """
 
 import re
-from typing import Mapping, Tuple, Callable, Any, Union, List
+from typing import Mapping, Tuple, Callable, Any, Union, List, Optional, Protocol
 
 from .utils import freeze
 
-__all__ = ('Query', 'where')
+__all__ = ('Query', 'QueryLike', 'where')
 
 
 def is_sequence(obj):
     return hasattr(obj, '__iter__')
 
 
+class QueryLike(Protocol):
+    """
+    A typing protocol that acts like a query.
+
+    Something that we use as a query must have two properties:
+
+    1. It must be callable, accepting a `Mapping` object and returning a
+       boolean that indicates whether the value matches the query, and
+    2. it must have a stable hash that will be used for query caching.
+
+    In addition, to mark a query as non-cacheable (e.g. if it involves
+    some remote lookup) it needs to have a method called ``is_cacheable``
+    that returns ``False``.
+
+    This query protocol is used to make MyPy correctly support the query
+    pattern that TinyDB uses.
+
+    See also https://mypy.readthedocs.io/en/stable/protocols.html#simple-user-defined-protocols
+    """
+    def __call__(self, value: Mapping) -> bool: ...
+
+    def __hash__(self) -> int: ...
+
+
 class QueryInstance:
     """
     A query instance.
@@ -45,23 +69,26 @@ class QueryInstance:
     instance can be used as a key in a dictionary.
     """
 
-    def __init__(self, test: Callable[[Mapping], bool], hashval: Tuple):
+    def __init__(self, test: Callable[[Mapping], bool], hashval: Optional[Tuple]):
         self._test = test
         self._hash = hashval
 
+    def is_cacheable(self) -> bool:
+        return self._hash is not None
+
     def __call__(self, value: Mapping) -> bool:
         """
         Evaluate the query to check if it matches a specified value.
 
         :param value: The value to check.
-        :return: Wether the value matchs this query.
+        :return: Whether the value matches this query.
         """
         return self._test(value)
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         # We calculate the query hash by using the ``hashval`` object which
-        # describes this query uniquely so we can calculate a stable hash value
-        # by simply hashing it
+        # describes this query uniquely, so we can calculate a stable hash
+        # value by simply hashing it
         return hash(self._hash)
 
     def __repr__(self):
@@ -79,26 +106,32 @@ class QueryInstance:
         # We use a frozenset for the hash as the AND operation is commutative
         # (a & b == b & a) and the frozenset does not consider the order of
         # elements
-        return QueryInstance(lambda value: self(value) and other(value),
-                             ('and', frozenset([self._hash, other._hash])))
+        if self.is_cacheable() and other.is_cacheable():
+            hashval = ('and', frozenset([self._hash, other._hash]))
+        else:
+            hashval = None
+        return QueryInstance(lambda value: self(value) and other(value), hashval)
 
     def __or__(self, other: 'QueryInstance') -> 'QueryInstance':
         # We use a frozenset for the hash as the OR operation is commutative
         # (a | b == b | a) and the frozenset does not consider the order of
         # elements
-        return QueryInstance(lambda value: self(value) or other(value),
-                             ('or', frozenset([self._hash, other._hash])))
+        if self.is_cacheable() and other.is_cacheable():
+            hashval = ('or', frozenset([self._hash, other._hash]))
+        else:
+            hashval = None
+        return QueryInstance(lambda value: self(value) or other(value), hashval)
 
     def __invert__(self) -> 'QueryInstance':
-        return QueryInstance(lambda value: not self(value),
-                             ('not', self._hash))
+        hashval = ('not', self._hash) if self.is_cacheable() else None
+        return QueryInstance(lambda value: not self(value), hashval)
 
 
 class Query(QueryInstance):
     """
     TinyDB Queries.
 
-    Allows to build queries for TinyDB databases. There are two main ways of
+    Allows building queries for TinyDB databases. There are two main ways of
     using queries:
 
     1) ORM-like usage:
@@ -124,12 +157,12 @@ class Query(QueryInstance):
 
     Queries are executed by calling the resulting object. They expect to get
     the document to test as the first argument and return ``True`` or
-    ``False`` depending on whether the documents matches the query or not.
+    ``False`` depending on whether the documents match the query or not.
     """
 
-    def __init__(self):
+    def __init__(self) -> None:
         # The current path of fields to access when evaluating the object
-        self._path = ()
+        self._path: Tuple[Union[str, Callable], ...] = ()
 
         # Prevent empty queries to be evaluated
         def notest(_):
@@ -156,18 +189,26 @@ class Query(QueryInstance):
         query._path = self._path + (item,)
 
         # ... and update the query hash
-        query._hash = ('path', query._path)
+        query._hash = ('path', query._path) if self.is_cacheable() else None
 
         return query
 
     def __getitem__(self, item: str):
         # A different syntax for ``__getattr__``
-        return getattr(self, item)
+
+        # We cannot call ``getattr(item)`` here as it would try to resolve
+        # the name as a method name first, only then call our ``__getattr__``
+        # method. By calling ``__getattr__`` directly, we make sure that
+        # calling e.g. ``Query()['test']`` will always generate a query for a
+        # document's ``test`` field instead of returning a reference to the
+        # ``Query.test`` method
+        return self.__getattr__(item)
 
     def _generate_test(
             self,
             test: Callable[[Any], bool],
             hashval: Tuple,
+            allow_empty_path: bool = False
     ) -> QueryInstance:
         """
         Generate a query based on a test function that first resolves the query
@@ -177,14 +218,17 @@ class Query(QueryInstance):
         :param hashval: The hash of the query.
         :return: A :class:`~tinydb.queries.QueryInstance` object
         """
-        if not self._path:
+        if not self._path and not allow_empty_path:
             raise ValueError('Query has no path')
 
         def runner(value):
             try:
                 # Resolve the path
                 for part in self._path:
-                    value = value[part]
+                    if isinstance(part, str):
+                        value = value[part]
+                    else:
+                        value = part(value)
             except (KeyError, TypeError):
                 return False
             else:
@@ -193,7 +237,7 @@ class Query(QueryInstance):
 
         return QueryInstance(
             lambda value: runner(value),
-            hashval
+            (hashval if self.is_cacheable() else None)
         )
 
     def __eq__(self, rhs: Any):
@@ -332,7 +376,7 @@ class Query(QueryInstance):
 
         .. warning::
 
-            The test fuction provided needs to be deterministic (returning the
+            The test function provided needs to be deterministic (returning the
             same value when provided with the same arguments), otherwise this
             may mess up the query cache that :class:`~tinydb.table.Table`
             implements.
@@ -427,6 +471,20 @@ class Query(QueryInstance):
             ('one_of', self._path, freeze(items))
         )
 
+    def fragment(self, document: Mapping) -> QueryInstance:
+        def test(value):
+            for key in document:
+                if key not in value or value[key] != document[key]:
+                    return False
+
+            return True
+
+        return self._generate_test(
+            lambda value: test(value),
+            ('fragment', freeze(document)),
+            allow_empty_path=True
+        )
+
     def noop(self) -> QueryInstance:
         """
         Always evaluate to ``True``.
@@ -439,6 +497,21 @@ class Query(QueryInstance):
             ()
         )
 
+    def map(self, fn: Callable[[Any], Any]) -> 'Query':
+        """
+        Add a function to the query path. Similar to __getattr__ but for
+        arbitrary functions.
+        """
+        query = type(self)()
+
+        # Now we add the callable to the query path ...
+        query._path = self._path + (fn,)
+
+        # ... and kill the hash - callable objects can be mutable, so it's
+        # harmful to cache their results.
+        query._hash = None
+
+        return query
 
 def where(key: str) -> Query:
     """
