@@ -12,11 +12,12 @@ from typing import (
     Mapping,
     Optional,
     Union,
-    cast
+    cast,
+    Tuple
 )
 
+from .queries import QueryLike
 from .storages import Storage
-from .queries import Query
 from .utils import LRUCache
 
 __all__ = ('Document', 'Table')
@@ -26,7 +27,7 @@ class Document(dict):
     """
     A document stored in the database.
 
-    This class provides a way to access both a document's content as well as
+    This class provides a way to access both a document's content and
     its ID using ``doc.doc_id``.
     """
 
@@ -49,7 +50,7 @@ class Table:
         once a threshold is reached.
 
         The query cache is updated on every search operation. When writing
-        data, the whole cache is discareded as the query results may have
+        data, the whole cache is discarded as the query results may have
         changed.
 
     .. admonition:: Customization
@@ -71,6 +72,7 @@ class Table:
     :param storage: The storage instance to use for this table
     :param name: The table name
     :param cache_size: Maximum capacity of query cache
+    :param persist_empty: Store new table even with no operations on it
     """
 
     #: The class used to represent documents
@@ -97,7 +99,8 @@ class Table:
         self,
         storage: Storage,
         name: str,
-        cache_size: int = default_query_cache_capacity
+        cache_size: int = default_query_cache_capacity,
+        persist_empty: bool = False
     ):
         """
         Create a table instance.
@@ -105,10 +108,12 @@ class Table:
 
         self._storage = storage
         self._name = name
-        self._query_cache = self.query_cache_class(capacity=cache_size) \
-            # type: LRUCache[Query, List[Document]]
+        self._query_cache: LRUCache[QueryLike, List[Document]] \
+            = self.query_cache_class(capacity=cache_size)
 
         self._next_id = None
+        if persist_empty:
+            self._update_table(lambda table: table.clear())
 
     def __repr__(self):
         args = [
@@ -146,7 +151,7 @@ class Table:
             raise ValueError('Document is not a Mapping')
 
         # First, we get the document ID for the new document
-        if isinstance(document, Document):
+        if isinstance(document, self.document_class):
             # For a `Document` object we use the specified ID
             doc_id = document.doc_id
 
@@ -159,7 +164,9 @@ class Table:
 
         # Now, we update the table and add the document
         def updater(table: dict):
-            assert doc_id not in table, 'doc_id '+str(doc_id)+' already exists'
+            if doc_id in table:
+                raise ValueError(f'Document with ID {str(doc_id)} '
+                                 f'already exists')
 
             # By calling ``dict(document)`` we convert the data we got to a
             # ``dict`` instance even if it was a different class that
@@ -175,24 +182,39 @@ class Table:
         """
         Insert multiple documents into the table.
 
-        :param documents: a Iterable of documents to insert
+        :param documents: an Iterable of documents to insert
         :returns: a list containing the inserted documents' IDs
         """
         doc_ids = []
 
         def updater(table: dict):
             for document in documents:
+
                 # Make sure the document implements the ``Mapping`` interface
                 if not isinstance(document, Mapping):
                     raise ValueError('Document is not a Mapping')
 
-                # Get the document ID for this document and store it so we
-                # can return all document IDs later
+                if isinstance(document, self.document_class):
+                    # Check if document does not override an existing document
+                    if document.doc_id in table:
+                        raise ValueError(
+                            f'Document with ID {str(document.doc_id)} '
+                            f'already exists'
+                        )
+
+                    # Store the doc_id, so we can return all document IDs
+                    # later. Then save the document with its doc_id and
+                    # skip the rest of the current loop
+                    doc_id = document.doc_id
+                    doc_ids.append(doc_id)
+                    table[doc_id] = dict(document)
+                    continue
+
+                # Generate new document ID for this document
+                # Store the doc_id, so we can return all document IDs
+                # later, then save the document with the new doc_id
                 doc_id = self._get_next_id()
                 doc_ids.append(doc_id)
-
-                # Convert the document to a ``dict`` (see Table.insert) and
-                # store it
                 table[doc_id] = dict(document)
 
         # See below for details on ``Table._update``
@@ -214,7 +236,7 @@ class Table:
 
         return list(iter(self))
 
-    def search(self, cond: Query) -> List[Document]:
+    def search(self, cond: QueryLike) -> List[Document]:
         """
         Search for all documents matching a 'where' cond.
 
@@ -224,39 +246,64 @@ class Table:
 
         # First, we check the query cache to see if it has results for this
         # query
-        if cond in self._query_cache:
-            docs = self._query_cache.get(cond)
-            if docs is not None:
-                return docs[:]
+        cached_results = self._query_cache.get(cond)
+        if cached_results is not None:
+            return cached_results[:]
 
-        # Perform the search by applying the query to all documents
-        docs = [doc for doc in self if cond(doc)]
+        # Perform the search by applying the query to all documents.
+        # Then, only if the document matches the query, convert it
+        # to the document class and document ID class.
+        docs = [
+            self.document_class(doc, self.document_id_class(doc_id))
+            for doc_id, doc in self._read_table().items()
+            if cond(doc)
+        ]
 
-        # Update the query cache
-        self._query_cache[cond] = docs[:]
+        # Only cache cacheable queries.
+        #
+        # This weird `getattr` dance is needed to make MyPy happy as
+        # it doesn't know that a query might have a `is_cacheable` method
+        # that is not declared in the `QueryLike` protocol due to it being
+        # optional.
+        # See: https://github.com/python/mypy/issues/1424
+        #
+        # Note also that by default we expect custom query objects to be
+        # cacheable (which means they need to have a stable hash value).
+        # This is to keep consistency with TinyDB's behavior before
+        # `is_cacheable` was introduced which assumed that all queries
+        # are cacheable.
+        is_cacheable: Callable[[], bool] = getattr(cond, 'is_cacheable',
+                                                   lambda: True)
+        if is_cacheable():
+            # Update the query cache
+            self._query_cache[cond] = docs[:]
 
         return docs
 
     def get(
         self,
-        cond: Optional[Query] = None,
+        cond: Optional[QueryLike] = None,
         doc_id: Optional[int] = None,
-    ) -> Optional[Document]:
+        doc_ids: Optional[List] = None
+    ) -> Optional[Union[Document, List[Document]]]:
         """
         Get exactly one document specified by a query or a document ID.
-
+        However, if multiple document IDs are given then returns all
+        documents in a list.
+        
         Returns ``None`` if the document doesn't exist.
 
         :param cond: the condition to check against
         :param doc_id: the document's ID
+        :param doc_ids: the document's IDs(multiple)
 
-        :returns: the document or ``None``
+        :returns: the document(s) or ``None``
         """
+        table = self._read_table()
 
         if doc_id is not None:
             # Retrieve a document specified by its ID
-            table = self._read_table()
-            raw_doc = table.get(doc_id, None)
+            raw_doc = table.get(str(doc_id), None)
 
             if raw_doc is None:
                 return None
@@ -264,19 +311,41 @@ class Table:
             # Convert the raw data to the document class
             return self.document_class(raw_doc, doc_id)
 
+        elif doc_ids is not None:
+            # Filter the table by extracting out all those documents which
+            # have doc id specified in the doc_id list.
+
+            # Since document IDs will be unique, we make it a set to ensure
+            # constant time lookup
+            doc_ids_set = set(str(doc_id) for doc_id in doc_ids)
+
+            # Now return the filtered documents in form of list
+            return [
+                self.document_class(doc, self.document_id_class(doc_id))
+                for doc_id, doc in table.items()
+                if doc_id in doc_ids_set
+            ]
+
         elif cond is not None:
             # Find a document specified by a query
-            for doc in self:
+            # The trailing underscore in doc_id_ is needed so MyPy
+            # doesn't think that `doc_id_` (which is a string) needs
+            # to have the same type as `doc_id` which is this function's
+            # parameter and is an optional `int`.
+            for doc_id_, doc in self._read_table().items():
                 if cond(doc):
-                    return doc
+                    return self.document_class(
+                        doc,
+                        self.document_id_class(doc_id_)
+                    )
 
             return None
 
-        raise RuntimeError('You have to pass either cond or doc_id')
+        raise RuntimeError('You have to pass either cond or doc_id or doc_ids')
 
     def contains(
         self,
-        cond: Optional[Query] = None,
+        cond: Optional[QueryLike] = None,
         doc_id: Optional[int] = None
     ) -> bool:
         """
@@ -301,7 +370,7 @@ class Table:
     def update(
         self,
         fields: Union[Mapping, Callable[[Mapping], None]],
-        cond: Optional[Query] = None,
+        cond: Optional[QueryLike] = None,
         doc_ids: Optional[Iterable[int]] = None,
     ) -> List[int]:
         """
@@ -348,7 +417,7 @@ class Table:
             updated_ids = []
 
             def updater(table: dict):
-                _cond = cast('Query', cond)
+                _cond = cast(QueryLike, cond)
 
                 # We need to convert the keys iterator to a list because
                 # we may remove entries from the ``table`` dict during
@@ -389,19 +458,90 @@ class Table:
 
             return updated_ids
 
-    def upsert(self, document: Mapping, cond: Query) -> List[int]:
+    def update_multiple(
+        self,
+        updates: Iterable[
+            Tuple[Union[Mapping, Callable[[Mapping], None]], QueryLike]
+        ],
+    ) -> List[int]:
         """
-        Update a document, if it exist, insert it otherwise.
+        Update all matching documents to have a given set of fields.
 
-        Note: this will update *all* documents matching the query.
-
-        :param document: the document to insert or the fields to update
-        :param cond: which document to look for
         :returns: a list containing the updated document's ID
         """
 
+        # Define the function that will perform the update
+        def perform_update(fields, table, doc_id):
+            if callable(fields):
+                # Update documents by calling the update function provided
+                # by the user
+                fields(table[doc_id])
+            else:
+                # Update documents by setting all fields from the provided
+                # data
+                table[doc_id].update(fields)
+
+        # Perform the update operation for documents specified by a query
+
+        # Collect affected doc_ids
+        updated_ids = []
+
+        def updater(table: dict):
+            # We need to convert the keys iterator to a list because
+            # we may remove entries from the ``table`` dict during
+            # iteration and doing this without the list conversion would
+            # result in an exception (RuntimeError: dictionary changed size
+            # during iteration)
+            for doc_id in list(table.keys()):
+                for fields, cond in updates:
+                    _cond = cast(QueryLike, cond)
+
+                    # Pass through all documents to find documents matching the
+                    # query. Call the processing callback with the document ID
+                    if _cond(table[doc_id]):
+                        # Add ID to list of updated documents
+                        updated_ids.append(doc_id)
+
+                        # Perform the update (see above)
+                        perform_update(fields, table, doc_id)
+
+        # Perform the update operation (see _update_table for details)
+        self._update_table(updater)
+
+        return updated_ids
+
+    def upsert(self, document: Mapping, cond: Optional[QueryLike] = None) -> List[int]:
+        """
+        Update documents, if they exist, insert them otherwise.
+
+        Note: This will update *all* documents matching the query. Document
+        argument can be a tinydb.table.Document object if you want to specify a
+        doc_id.
+
+        :param document: the document to insert or the fields to update
+        :param cond: which document to look for, optional if you've passed a
+        Document with a doc_id
+        :returns: a list containing the updated documents' IDs
+        """
+
+        # Extract doc_id
+        if isinstance(document, self.document_class) and hasattr(document, 'doc_id'):
+            doc_ids: Optional[List[int]] = [document.doc_id]
+        else:
+            doc_ids = None
+
+        # Make sure we can actually find a matching document
+        if doc_ids is None and cond is None:
+            raise ValueError("If you don't specify a search query, you must "
+                             "specify a doc_id. Hint: use a table.Document "
+                             "object.")
+
         # Perform the update operation
-        updated_docs = self.update(document, cond)
+        try:
+            updated_docs: Optional[List[int]] = self.update(document, cond, doc_ids)
+        except KeyError:
+            # This happens when a doc_id is specified, but it's missing
+            updated_docs = None
 
         # If documents have been updated: return their IDs
         if updated_docs:
@@ -413,7 +553,7 @@ class Table:
 
     def remove(
         self,
-        cond: Optional[Query] = None,
+        cond: Optional[QueryLike] = None,
         doc_ids: Optional[Iterable[int]] = None,
     ) -> List[int]:
         """
@@ -423,8 +563,24 @@ class Table:
         :param doc_ids: a list of document IDs
         :returns: a list containing the removed documents' ID
         """
-        if cond is None and doc_ids is None:
-            raise RuntimeError('Use truncate() to remove all documents')
+        if doc_ids is not None:
+            # This function returns the list of IDs for the documents that have
+            # been removed. When removing documents identified by a set of
+            # document IDs, it's this list of document IDs we need to return
+            # later.
+            # We convert the document ID iterator into a list, so we can both
+            # use the document IDs to remove the specified documents and
+            # to return the list of affected document IDs
+            removed_ids = list(doc_ids)
+
+            def updater(table: dict):
+                for doc_id in removed_ids:
+                    table.pop(doc_id)
+
+            # Perform the remove operation
+            self._update_table(updater)
+
+            return removed_ids
 
         if cond is not None:
             removed_ids = []
@@ -436,7 +592,7 @@ class Table:
                 # We need to convince MyPy (the static type checker) that
                 # the ``cond is not None`` invariant still holds true when
                 # the updater function is called
-                _cond = cast('Query', cond)
+                _cond = cast(QueryLike, cond)
 
                 # We need to convert the keys iterator to a list because
                 # we may remove entries from the ``table`` dict during
@@ -456,26 +612,7 @@ class Table:
 
             return removed_ids
 
-        if doc_ids is not None:
-            # This function returns the list of IDs for the documents that have
-            # been removed. When removing documents identified by a set of
-            # document IDs, it's this list of document IDs we need to return
-            # later.
-            # We convert the document ID iterator into a list so we can both
-            # use the document IDs to remove the specified documents as well as
-            # to return the list of affected document IDs
-            removed_ids = list(doc_ids)
-
-            def updater(table: dict):
-                for doc_id in removed_ids:
-                    table.pop(doc_id)
-
-            # Perform the remove operation
-            self._update_table(updater)
-
-            return removed_ids
-
-        raise RuntimeError('This should never happen')
+        raise RuntimeError('Use truncate() to remove all documents')
 
     def truncate(self) -> None:
         """
@@ -488,7 +625,7 @@ class Table:
         # Reset document ID counter
         self._next_id = None
 
-    def count(self, cond: Query) -> int:
+    def count(self, cond: QueryLike) -> int:
         """
         Count the documents matching a query.
 
@@ -509,20 +646,7 @@ class Table:
         Count the total number of documents in this table.
         """
 
-        # Using self._read_table() will convert all documents into
-        # the document class. But for counting the number of documents
-        # this conversion is not necessary, thus we read the storage
-        # directly here
-
-        tables = self._storage.read()
-
-        if tables is None:
-            return 0
-
-        try:
-            return len(tables[self.name])
-        except KeyError:
-            return 0
+        return len(self._read_table())
 
     def __iter__(self) -> Iterator[Document]:
         """
@@ -534,7 +658,7 @@ class Table:
         # Iterate all documents and their IDs
         for doc_id, doc in self._read_table().items():
             # Convert documents to the document class
-            yield self.document_class(doc, doc_id)
+            yield self.document_class(doc, self.document_id_class(doc_id))
 
     def _get_next_id(self):
         """
@@ -565,20 +689,19 @@ class Table:
         max_id = max(self.document_id_class(i) for i in table.keys())
         next_id = max_id + 1
 
-        # The next ID we wil return AFTER this call needs to be larger than
+        # The next ID we will return AFTER this call needs to be larger than
         # the current next ID we calculated
         self._next_id = next_id + 1
 
         return next_id
 
-    def _read_table(self) -> Dict[int, Mapping]:
+    def _read_table(self) -> Dict[str, Mapping]:
         """
         Read the table data from the underlying storage.
 
-        Here we read the data from the underlying storage and convert all
-        IDs to the document ID class. Documents themselves are NOT yet
-        transformed into the document class, we may not want to convert
-        *all* documents when returning only one document for example.
+        Documents and doc_ids are NOT yet transformed, as
+        we may not want to convert *all* documents when returning
+        only one document for example.
         """
 
         # Retrieve the tables from the storage
@@ -595,19 +718,14 @@ class Table:
             # The table does not exist yet, so it is empty
             return {}
 
-        # Convert all document IDs to the correct document ID class and return
-        # the table data dict
-        return {
-            self.document_id_class(doc_id): doc
-            for doc_id, doc in table.items()
-        }
+        return table
 
     def _update_table(self, updater: Callable[[Dict[int, Mapping]], None]):
         """
-        Perform an table update operation.
+        Perform a table update operation.
 
         The storage interface used by TinyDB only allows to read/write the
-        complete database data, but not modifying only portions of it. Thus
+        complete database data, but not modifying only portions of it. Thus,
         to only update portions of the table data, we first perform a read
         operation, perform the update on the table data and then write
         the updated data back to the storage.
@@ -642,7 +760,7 @@ class Table:
 
         # Convert the document IDs back to strings.
         # This is required as some storages (most notably the JSON file format)
-        # don't require IDs other than strings.
+        # don't support IDs other than strings.
         tables[self.name] = {
             str(doc_id): doc
             for doc_id, doc in table.items()
